import math
import torch
from torch import nn, optim, Tensor
from typing import Optional
import json
import random
from transformers import AutoTokenizer

# DO NOT CHANGE THIS LINE!
# And DO NOT reset the torch seed anywhere else in your code!
torch.manual_seed(10601)

class RNNCell(nn.Module):
    def __init__(self, input_dim, hidden_dim, nonlinearity='tanh'):
        """
        input_dim (int): Input dimension of RNN
        hidden_dim (int): Hidden dimension of RNN
        nonlinearity (str): Nonlinearity in RNN (either 'tanh' or 'relu')
        """
        super(RNNCell, self).__init__()
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim

        # These are just linear transformations, so you can use nn.Linear
        self.i2h = # TODO: fill this in
        self.h2h = # TODO: fill this in

        # See here for PyTorch activation functions
        # https://pytorch.org/docs/stable/nn.html#non-linear-activations-weighted-sum-nonlinearity
        if nonlinearity == 'tanh':
            self.activation = # TODO: fill this in
        elif nonlinearity == 'relu':
            self.activation = # TODO: fill this in
        else:
            raise Exception(f"unknown nonlinearity: {nonlinearity}")

    def forward(self, input: Tensor, hidden_state: Tensor) -> Tensor:
        """
        Args:
            input (Tensor): Input at timestep t
                - shape: (input_dim,)
            hidden_state (Tensor): Hidden state from timestep t-1
                - shape: (hidden_dim,)

        Returns:
            Tensor: Next hidden state at timestep t
                - shape: (hidden_dim,)
        """

        return # TODO: fill this in


class RNN(nn.Module):
    def __init__(self, input_dim: int, hidden_dim: int, nonlinearity='tanh', use_attention=False):
        """
        input_dim (int): Input dimension of RNN
        hidden_dim (int): Hidden dimension of RNN
        nonlinearity (str): Nonlinearity in RNN (either 'tanh' or 'relu')
        use_attention (bool): Whether attention should be used in the RNN
        """
        super(RNN, self).__init__()
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        # use your RNNCell class!
        self.cell = # TODO: fill this in

        self.attention = None
        # only create attention module if we're using attention
        if use_attention:
            # use your SelfAttention class!
            # set the attention key_dim and value_dim to be the same as hidden_dim 
            self.attention = # TODO: fill this in


    def step(self, input: Tensor, hidden_states: Optional[Tensor] = None) -> Tensor:
        """
        Compute hidden and output states for a single timestep

        Args:
            input (Tensor): input at current timestep t 
                - shape: (input_dim,)
            hidden_states (Tensor): hidden states of preceding timesteps [1, t-1]
                If there are no previous hidden states (i.e. we are at t=1), then
                this may be None and we will initialize the previous hidden state 
                to all zeros.
                - shape: (t-1, hidden_dim)
        
        Returns:
            Tensor: RNN hidden state at current timestep t 
                - shape: (hidden_dim,)
            Tensor: RNN output at current timestep t. 
                When attention is not used, the output is the same as the RNN hidden state.
                When attention is used, output is the attention output (plus hidden
                state residual).
                - shape: (hidden_dim,)
        """
        if hidden_states is None:
            # if this is the first timestep and there is no previous hidden state,
            # create a dummy hidden state of all zeros
            last_hidden_state = # TODO: fill this in
        else:
            last_hidden_state = # TODO: fill this in

        next_hidden_state = next_output_state = self.cell(input, last_hidden_state)

        if self.attention is not None:
            if hidden_states is None:
                hidden_states = next_hidden_state.unsqueeze(0)
            else:
                hidden_states = torch.cat([hidden_states, next_hidden_state.unsqueeze(0)])
            next_output_state, _ = self.attention(hidden_states)
            
            # add residual since this works better in practice
            next_output_state = next_output_state + next_hidden_state

        return next_hidden_state, next_output_state

    def forward(self, sequence: Tensor) -> Tensor:
        """
        Compute hidden and output states for all timesteps over input sequence

        Args:
            sequence (Tensor): inputs to RNN over t timesteps 
                - shape (t, input_dim)
        
        Returns:
            Tensor: hidden states over t timesteps 
                - shape (t, hidden_dim)
            Tensor: output states over t timesteps 
                - shape (t, hidden_dim)
        """
        hidden_states = None
        output_states = []

        for input in sequence:
            # Compute next hidden state and next output state
            next_hidden_state, next_output_state = # TODO: fill this in
            
            # Update hidden_states and output_states
            # TODO: fill this in

        return hidden_states, output_states


class SelfAttention(nn.Module):
    '''Scaled dot product attention from original transformers paper'''

    def __init__(self, hidden_dim, key_dim, value_dim):
        """
        hidden_dim (int): Hidden dimension of RNN
        key_dim (int): Dimension of attention key and query vectors
        value_dim (int): Dimension of attention value vectors
        """
        super(SelfAttention, self).__init__()
        self.hidden_dim = hidden_dim
        self.key_dim = key_dim
        self.value_dim = value_dim

        self.query_transform = # TODO: fill this in
        self.key_transform = # TODO: fill this in
        self.value_transform = # TODO: fill this in

        self.output_transform = nn.Linear(value_dim, hidden_dim)

    def forward(self, hidden_states: Tensor) -> Tensor:
        """
        Compute attention for **current** timestep t

        Args:
            sequence (Tensor): Hidden states up to current timestep t 
                - shape (t, hidden_dim)

        Returns:
            Tensor: Attention output for current timestep 
                - shape (hidden_dim,)
            Tensor: Attention weights over values 
                - shape (t,)
        """
        last_hidden_state = hidden_states[-1]
        query = # TODO: fill this in
        keys = # TODO: fill this in
        values = # TODO: fill this in

        scaling = self.key_dim ** (0.5)

        # Compute attention weights over values
        # Remember to divide raw attention scores by scaling factor
        # These scores should then be normalized using softmax
        # Hint: use torch.softmax
        weights = # TODO: fill this in

        # Compute weighted sum of values based on attention weights
        output_state = # TODO: fill this in

        # Apply output projection back to hidden dimension
        output_state = self.output_transform(output_state)

        return output_state, weights


class RNNLanguageModel(nn.Module):
    def __init__(self, embed_dim, hidden_dim, vocab_size, nonlinearity='tanh', use_attention=False):
        """
        embed_dim (int): Dimension of word embeddings
        hidden_dim (int): Dimension of RNN hidden states
        vocab_size (int): Number of (sub)words in model vocabulary
        nonlinearity (str): Nonlinearity in RNN (either 'tanh' or 'relu')
        use_attention (bool): Whether to apply attention in RNN
        """
        super(RNNLanguageModel, self).__init__()

        # Word embeddings
        # Hint: use nn.Embedding
        self.embeddings = # TODO: fill this in

        # RNN backbone
        # use your RNN class!
        self.rnn = # TODO: fill this in
        
        # Final linear projection from RNN output state to next token logits
        self.lm_head = # TODO: fill this in
        
    def forward(self, tokens: Tensor) -> Tensor:
        """
        Computes next-token logits and hidden states for each token in tokens

        Args:
            tokens (Tensor): Input tokens IDs 
                - shape (t,)
        
        Returns:
            Tensor: Next-token logits for each token
                - shape (t, vocab_size)
            Tensor: Hidden states for each token
                - shape (t, hidden_dim)
        """
        # apply embeddings, rnns, and lm_head sequentially
        return # TODO: fill this in

    def select_token(self, token_logits: Tensor, temperature: float) -> int:
        """
        Selects (or samples) next token from token_logits

        Args:
            token_logits (Tensor): Next token logits 
                - shape (vocab_size,)
            temperature (float): Sampling temperature. If 0, do greedy decoding.

        Returns:
            int: ID of next token selected
        """
        if temperature == 0: # greedy decoding
            return # TODO: fill this in
        else: # temperature sampling
            # Hint: use torch.multinomial
            return # TODO: fill this in

    def generate(self, tokens: Tensor, max_tokens=10, temperature=0.0) -> Tensor:
        """
        Generates new tokens given `tokens` as a prefix.

        Args:
            tokens (Tensor): Input tokens 
                - shape: (input_length,)
            max_tokens (int): Number of new tokens to generate
            temperature (float): Sampling temperature

        Returns:
            Tensor: generated tokens
                - shape: (max_tokens,)
        """
        # get hidden states for input tokens
        token_logits, hidden_states = self(tokens)

        next_token_logits = token_logits[-1]
        new_tokens = []
        step = 0
        # now, start generating new tokens
        # NOTE: while you could in theory repeatedly call self(tokens) here, you shouldn't since
        # that's an order of magnitude more inefficient as you would be repeatedly re-encoding
        # the prefix. Instead, here, we repeatedly compute the hidden state and next token
        # logits for the *latest* token.
        while True:
            step += 1
            # select next token based on next_token_logits
            next_token = self.select_token(next_token_logits, temperature)
            new_tokens.append(next_token)

            if step >= max_tokens:
                break

            # get next input embedding (hint: use `next_token`)
            embed = # TODO: fill this in
            
            # get *next* hidden state and *next* output state from RNN
            next_hidden_state, next_output_state = # TODO: fill this in
            
            # get next token logits
            next_token_logits = # TODO: fill this in
            
            # update hidden states
            hidden_states = torch.cat([hidden_states, next_hidden_state.unsqueeze(0)])
            
        return torch.tensor(new_tokens)


def shuffle_dataset(dataset, seed):
    """
    Makes a shuffled copy of the dataset

    DO NOT MODIFY THIS FUNCTION.
    """
    random.seed(seed)
    indices = list(range(len(dataset)))
    random.shuffle(indices)
    dataset = [dataset[i] for i in indices]
    return dataset


def train(lm, dataset, loss_fn, optimizer, epoch):
    """
    Run one epoch of language model training

    Args:
        lm (RNNLanguageModel): RNN language model
        dataset (list[Tensor]): Train dataset
        loss_fn: PyTorch cross entropy loss function
        optimizer: PyTorch Adam optimizer
        epoch (int): Current epoch. Used as seed for shuffling dataset

    Returns:
        float: Average training loss
    """
    # Set the model to training model
    lm.train()
    max_grad_norm = 1.0
    mean_loss = 0.0

    # DO NOT change the next line
    dataset = shuffle_dataset(dataset, epoch)

    for idx, sequence in enumerate(dataset):
        # Forward pass through model
        token_logits, hidden_states = # TODO: fill this in

        # Compute next-token classification loss
        loss = # TODO: fill this in

        # Backward pass through model
        # TODO: fill this in

        # DO NOT change this - clip gradient norm to avoid exploding gradients
        nn.utils.clip_grad_norm_(lm.parameters(), max_grad_norm)

        # Update weights
        # TODO: fill this line

        # Zero gradients
        # TODO: fill this in

        mean_loss += loss.detach().cpu().item()

    return mean_loss / len(dataset)


@torch.no_grad()
def validate(lm, dataset, loss_fn):
    """
    Args:
        lm (RNNLanguageModel):
        dataset (list[Tensor]): Validation dataset
        loss_fn:

    Returns:
        float: Average validation loss
    """
    # Set the model to eval mode
    lm.eval()
    mean_loss = 0.0
    for sequence in dataset:
        # Forward pass through model
        token_logits, hidden_states = # TODO: fill this in

        # Compute loss
        loss = # TODO: fill this in

        mean_loss += loss.detach().cpu().item()
    return mean_loss / len(dataset)


@torch.no_grad()
def complete(prefix: str, num_tokens=64, temperature=0.0):
    """
    Generates text completion from language model given text prefix. 
    This function has been implemented for you.

    Args:
        prefix (str):
        num_tokens (int): Number of new tokens to generate
        temperature (float): Sampling temperature

    Returns:
        str: Text completion
    """
    lm.eval()
    input = tokenizer.encode(prefix, add_special_tokens=False, return_tensors='pt').squeeze(0)
    input = input.to(device)
    output = lm.generate(input, max_tokens=num_tokens, temperature=temperature)
    return tokenizer.decode(output)


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()

    parser.add_argument("--train_data", type=str)
    parser.add_argument("--val_data", type=str)
    parser.add_argument("--embed_dim", type=int)
    parser.add_argument("--hidden_dim", type=int)
    parser.add_argument("--activation", type=str)
    parser.add_argument("--use_attention", type=int)
    parser.add_argument("--lr", type=float)
    parser.add_argument("--num_epochs", type=int)
    parser.add_argument("--metrics_out", type=str)

    args = parser.parse_args()

    # initialize torch device to use cuda if we have a gpu
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    # load tokenizer
    tokenizer = AutoTokenizer.from_pretrained("my_tokenizer")
    vocab_size = tokenizer.vocab_size

    # initialize LM
    lm = RNNLanguageModel(
        embed_dim=args.embed_dim, 
        hidden_dim=args.hidden_dim, 
        vocab_size=vocab_size, 
        nonlinearity=args.activation, 
        use_attention=args.use_attention
    )
    lm = lm.to(device)
    print(lm)

    with open(args.train_data) as f:
        train_data = json.load(f)
    train_data = [torch.tensor(seq).to(device) for seq in train_data]

    with open(args.val_data) as f:
        valid_data = json.load(f)
    valid_data = [torch.tensor(seq).to(device) for seq in valid_data]

    loss_fn = nn.CrossEntropyLoss()
    optimizer = optim.Adam(lm.parameters(), lr=args.lr)

    train_losses = []
    valid_losses = []

    for epoch in range(args.num_epochs):
        epoch_str = str(epoch).rjust(2)

        train_loss = train(lm, train_data, loss_fn, optimizer, epoch)
        print(f"Epoch [{epoch_str}] Train loss = {train_loss}")

        valid_loss = validate(lm, valid_data, loss_fn)
        print(f"Epoch [{epoch_str}] Valid loss = {valid_loss}")

        train_losses.append(train_loss)
        valid_losses.append(valid_loss)

    # Saves your trained model weights
    torch.save(lm, "model.pt")

    # You can later load back in your model in a separate Python file by running:
    # >>> from rnn import *
    # >>> lm = torch.load("model.pt")
    # This may be helpful for Empirical Question 3.3, where
    # training the model may take up to 30 minutes.

    """
    # Example code for generating text with your LM

    test_str = ["Once upon a time there was a"]
    
    for ts in test_str:
        completion = complete(ts, num_tokens=64, temperature=0.3)
        print("  Test prefix:", ts)
        print("  Test output:", completion)
    """

    with open(args.metrics_out, "w") as f:
        for i in range(len(train_losses)):
            cur_epoch = i + 1
            cur_tr_loss = train_losses[i]
            cur_val_loss = valid_losses[i]
            f.write("epoch={} crossentropy(train): {}\n".format(
                cur_epoch, cur_tr_loss))
            f.write("epoch={} crossentropy(validation): {}\n".format(
                cur_epoch, cur_val_loss))
